@@ -2,14 +2,12 @@
 shapefile.py
 Provides read and write support for ESRI Shapefiles.
 author: jlawhead<at>geospatialpython.com
-date: 2015/06/22
-version: 1.2.8
-Compatible with Python versions 2.4-3.x
-version changelog: 
-- Pushing 1.2.5 changes to PyPi using Travis-CI.
+date: 2017/04/29
+version: 1.2.11
+Compatible with Python versions 2.7-3.x
 """
 
-__version__ = "1.2.8"
+__version__ = "1.2.11"
 
 from struct import pack, unpack, calcsize, error, Struct
 import os
@@ -18,6 +16,7 @@ import time
 import array
 import tempfile
 import itertools
+from datetime import date
 
 #
 # Constants for shape types
@@ -36,6 +35,8 @@ POLYGONM = 25
 MULTIPOINTM = 28
 MULTIPATCH = 31
 
+MISSING = [None,'']
+
 PYTHON3 = sys.version_info[0] == 3
 
 if PYTHON3:
@@ -105,7 +106,7 @@ def signed_area(coords):
     return sum(xs[i]*(ys[i+1]-ys[i-1]) for i in range(1, len(coords)))/2.0
 
 class _Shape:
-    def __init__(self, shapeType=None):
+    def __init__(self, shapeType=NULL):
         """Stores the geometry of the different shape types
         specified in the Shapefile spec. Shape types are
         usually point, polyline, or polygons. Every shape type
@@ -117,6 +118,7 @@ class _Shape:
         list of shapes."""
         self.shapeType = shapeType
         self.points = []
+        self.parts = []
 
     @property
     def __geo_interface__(self):
@@ -256,23 +258,24 @@ class Reader:
     def load(self, shapefile=None):
         """Opens a shapefile from a filename or file-like
         object. Normally this method would be called by the
-        constructor with the file object or file name as an
-        argument."""
+        constructor with the file name as an argument."""
         if shapefile:
             (shapeName, ext) = os.path.splitext(shapefile)
             self.shapeName = shapeName
             try:
                 self.shp = open("%s.shp" % shapeName, "rb")
             except IOError:
-                raise ShapefileException("Unable to open %s.shp" % shapeName)
+                pass
             try:
                 self.shx = open("%s.shx" % shapeName, "rb")
             except IOError:
-                raise ShapefileException("Unable to open %s.shx" % shapeName)
+                pass
             try:
                 self.dbf = open("%s.dbf" % shapeName, "rb")
             except IOError:
-                raise ShapefileException("Unable to open %s.dbf" % shapeName)
+                pass
+            if not (self.shp and self.dbf):
+                raise ShapefileException("Unable to open %s.dbf or %s.shp." % (shapeName, shapeName) )
         if self.shp:
             self.__shpHeader()
         if self.dbf:
@@ -439,23 +442,16 @@ class Reader:
         while shp.tell() < self.shpLength:
             yield self.__shape()    
 
-    def __dbfHeaderLength(self):
-        """Retrieves the header length of a dbf file header."""
-        if not self.__dbfHdrLength:
-            if not self.dbf:
-                raise ShapefileException("Shapefile Reader requires a shapefile or file-like object. (no dbf file found)")
-            dbf = self.dbf
-            (self.numRecords, self.__dbfHdrLength) = \
-                    unpack("<xxxxLH22x", dbf.read(32))
-        return self.__dbfHdrLength
-
     def __dbfHeader(self):
         """Reads a dbf header. Xbase-related code borrows heavily from ActiveState Python Cookbook Recipe 362715 by Raymond Hettinger"""
         if not self.dbf:
             raise ShapefileException("Shapefile Reader requires a shapefile or file-like object. (no dbf file found)")
         dbf = self.dbf
-        headerLength = self.__dbfHeaderLength()
-        numFields = (headerLength - 33) // 32
+        # read relevant header parts
+        self.numRecords, self.__dbfHdrLength, self.__recordLength = \
+                unpack("<xxxxLHH20x", dbf.read(32))
+        # read fields
+        numFields = (self.__dbfHdrLength - 33) // 32
         for field in range(numFields):
             fieldDesc = list(unpack("<11sc4xBB14x", dbf.read(32)))
             name = 0
@@ -477,11 +473,16 @@ class Reader:
         self.__recStruct = Struct(fmt)
 
     def __recordFmt(self):
-        """Calculates the size of a .shp geometry record."""
-        if not self.numRecords:
+        """Calculates the format and size of a .dbf record."""
+        if self.numRecords is None:
             self.__dbfHeader()
         fmt = ''.join(['%ds' % fieldinfo[2] for fieldinfo in self.fields])
         fmtSize = calcsize(fmt)
+        # total size of fields should add up to recordlength from the header
+        while fmtSize < self.__recordLength:
+            # if not, pad byte until reaches recordlength
+            fmt += "x" 
+            fmtSize += 1
         return (fmt, fmtSize)
 
     def __record(self):
@@ -492,14 +493,11 @@ class Reader:
             # deleted record
             return None
         record = []
-        for (name, typ, size, deci), value in zip(self.fields,
-                                                                                                recordContents):
+        for (name, typ, size, deci), value in zip(self.fields, recordContents):
             if name == 'DeletionFlag':
                 continue
-            elif not value.strip():
-                record.append(value)
-                continue
-            elif typ == "N":
+            elif typ in ("N","F"):
+                # numeric or float: number stored as a string, right justified, and padded with blanks to the width of the field. 
                 value = value.replace(b('\0'), b('')).strip()
                 value = value.replace(b('*'), b(''))  # QGIS NULL is all '*' chars
                 if value == b(''):
@@ -516,19 +514,29 @@ class Reader:
                     except ValueError:
                         #not parseable as int, set to None
                         value = None
-            elif typ == b('D'):
+            elif typ == 'D':
+                # date: 8 bytes - date stored as a string in the format YYYYMMDD.
                 if value.count(b('0')) == len(value):  # QGIS NULL is all '0' chars
                     value = None
                 else:
                     try:
                         y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
-                        value = [y, m, d]
+                        value = date(y, m, d)
                     except:
                         value = value.strip()
-            elif typ == b('L'):
-                value = (value in b('YyTt') and b('T')) or \
-                                        (value in b('NnFf') and b('F')) or b('?')
+            elif typ == 'L':
+                # logical: 1 byte - initialized to 0x20 (space) otherwise T or F.
+                if value == b(" "):
+                    value = None # space means missing or not yet set
+                else:
+                    if value in b('YyTt1'):
+                        value = True
+                    elif value in b('NnFf0'):
+                        value = False
+                    else:
+                        value = None # unknown value is set to missing
             else:
+                # anything else is forced to string/unicode
                 value = u(value)
                 value = value.strip()
             record.append(value)
@@ -537,32 +545,34 @@ class Reader:
     def record(self, i=0):
         """Returns a specific dbf record based on the supplied index."""
         f = self.__getFileObj(self.dbf)
-        if not self.numRecords:
+        if self.numRecords is None:
             self.__dbfHeader()
         i = self.__restrictIndex(i)
         recSize = self.__recStruct.size
         f.seek(0)
-        f.seek(self.__dbfHeaderLength() + (i * recSize))
+        f.seek(self.__dbfHdrLength + (i * recSize))
         return self.__record()
 
     def records(self):
         """Returns all records in a dbf file."""
-        if not self.numRecords:
+        if self.numRecords is None:
             self.__dbfHeader()
+        records = []
         f = self.__getFileObj(self.dbf)
-        f.seek(self.__dbfHeaderLength())
-        flat = unpack(self.__recStruct.format * self.numRecords, f.read(self.__recStruct.size * self.numRecords))
-        rowlen = len(self.fields) - 1
-        records = list(izip(*(iter(flat),) * rowlen))
+        f.seek(self.__dbfHdrLength)
+        for i in range(self.numRecords):
+            r = self.__record()
+            if r:
+                records.append(r)
         return records
 
     def iterRecords(self):
         """Serves up records in a dbf file as an iterator.
         Useful for large shapefiles or dbf files."""
-        if not self.numRecords:
+        if self.numRecords is None:
             self.__dbfHeader()
         f = self.__getFileObj(self.dbf)
-        f.seek(self.__dbfHeaderLength())
+        f.seek(self.__dbfHdrLength)
         for i in xrange(self.numRecords):
             r = self.__record()
             if r:
@@ -657,7 +667,7 @@ class Writer:
                 # z array
                 size += 8 * nPoints
             # Calc m extremes and values
-            if self.shapeType in (23,25,31):
+            if self.shapeType in (15,23,25,31):
                 # m extremes
                 size += 16
                 # m array
@@ -675,19 +685,19 @@ class Writer:
         size //= 2
         return size
 
-    def __bbox(self, shapes, shapeTypes=[]):
+    def __bbox(self, shapes):
         x = []
         y = []
         for s in shapes:
-            shapeType = self.shapeType
-            if shapeTypes:
-                shapeType = shapeTypes[shapes.index(s)]
-            px, py = list(zip(*s.points))[:2]
-            x.extend(px)
-            y.extend(py)
+            if len(s.points) > 0:
+                px, py = list(zip(*s.points))[:2]
+                x.extend(px)
+                y.extend(py)
+        if len(x) == 0:
+            return [0] * 4
         return [min(x), min(y), max(x), max(y)]
 
-    def __zbox(self, shapes, shapeTypes=[]):
+    def __zbox(self, shapes):
         z = []
         for s in shapes:
             try:
@@ -698,14 +708,15 @@ class Writer:
         if not z: z.append(0)
         return [min(z), max(z)]
 
-    def __mbox(self, shapes, shapeTypes=[]):
-        m = [0]
+    def __mbox(self, shapes):
+        m = []
         for s in shapes:
             try:
                 for p in s.points:
                     m.append(p[3])
             except IndexError:
                 pass
+        if not m: m.append(0)
         return [min(m), max(m)]
 
     def bbox(self):
@@ -763,7 +774,7 @@ class Writer:
         year -= 1900
         # Remove deletion flag placeholder from fields
         for field in self.fields:
-            if field[0].startswith("Deletion"):
+            if str(field[0]).startswith("Deletion"):
                 self.fields.remove(field)
         numRecs = len(self.records)
         numFields = len(self.fields)
@@ -797,8 +808,6 @@ class Writer:
             recNum += 1
             start = f.tell()
             # Shape Type
-            if self.shapeType != 31:
-                s.shapeType = self.shapeType
             f.write(pack("<i", s.shapeType))
             # All shape types capable of having a bounding box
             if s.shapeType in (3,5,8,13,15,18,23,25,28,31):
@@ -844,14 +853,14 @@ class Writer:
             # Write m extremes and values
             if s.shapeType in (13,15,18,23,25,28,31):
                 try:
-                    if hasattr(s,"m"):
+                    if hasattr(s,"m") and None not in s.m:
                         f.write(pack("<%sd" % len(s.m), *s.m))
                     else:
                         f.write(pack("<2d", *self.__mbox([s])))
                 except error:
                     raise ShapefileException("Failed to write measure extremes for record %s. Expected floats" % recNum)
                 try:
-                    [f.write(pack("<d", p[3])) for p in s.points]
+                    [f.write(pack("<d", len(p) > 3 and p[3] or 0)) for p in s.points]
                 except error:
                     raise ShapefileException("Failed to write measure values for record %s. Expected floats" % recNum)
             # Write a single point
@@ -915,14 +924,41 @@ class Writer:
         for record in self.records:
             if not self.fields[0][0].startswith("Deletion"):
                 f.write(b(' ')) # deletion flag
-            for (fieldName, fieldType, size, dec), value in zip(self.fields, record):
+            for (fieldName, fieldType, size, deci), value in zip(self.fields, record):
                 fieldType = fieldType.upper()
                 size = int(size)
-                if fieldType.upper() == "N":
-                    value = str(value).rjust(size)
+                if fieldType in ("N","F"):
+                    # numeric or float: number stored as a string, right justified, and padded with blanks to the width of the field.
+                    if value in MISSING:
+                        value = str("*"*size) # QGIS NULL
+                    elif not deci:
+                        value = format(value, "d")[:size].rjust(size) # caps the size if exceeds the field size
+                    else:
+                        value = format(value, ".%sf"%deci)[:size].rjust(size) # caps the size if exceeds the field size
+                elif fieldType == "D":
+                    # date: 8 bytes - date stored as a string in the format YYYYMMDD.
+                    if isinstance(value, date):
+                        value = value.strftime("%Y%m%d")
+                    elif isinstance(value, list) and len(value) == 3:
+                        value = date(*value).strftime("%Y%m%d")
+                    elif value in MISSING:
+                        value = b('0') * 8 # QGIS NULL for date type
+                    elif isinstance(value, str) and len(value) == 8:
+                        pass # value is already a date string
+                    else:
+                        raise ShapefileException("Date values must be either a datetime.date object, a list, a YYYYMMDD string, or a missing value.")
                 elif fieldType == 'L':
-                    value = str(value)[0].upper()
+                    # logical: 1 byte - initialized to 0x20 (space) otherwise T or F.
+                    if value in MISSING:
+                        value = b(' ') # missing is set to space
+                    elif value in [True,1]:
+                        value = b("T")
+                    elif value in [False,0]:
+                        value = b("F")
+                    else:
+                        value = b(' ') # unknown is set to space
                 else:
+                    # anything else is forced to string
                     value = str(value)[:size].ljust(size)
                 if len(value) != size:
                     raise ShapefileException(
@@ -935,9 +971,9 @@ class Writer:
         """Creates a null shape."""
         self._shapes.append(_Shape(NULL))
 
-    def point(self, x, y, z=0, m=0):
+    def point(self, x, y, z=0, m=0, shapeType=POINT):
         """Creates a point shape."""
-        pointShape = _Shape(self.shapeType)
+        pointShape = _Shape(shapeType)
         pointShape.points.append([x, y, z, m])
         self._shapes.append(pointShape)
 
@@ -980,6 +1016,12 @@ class Writer:
 
     def field(self, name, fieldType="C", size="50", decimal=0):
         """Adds a dbf field descriptor to the shapefile."""
+        if fieldType == "D":
+            size = "8"
+            decimal = 0
+        elif fieldType == "L":
+            size = "1"
+            decimal = 0
         self.fields.append((name, fieldType, size, decimal))
 
     def record(self, *recordList, **recordDict):
@@ -995,7 +1037,7 @@ class Writer:
         # Compensate for deletion flag
         if self.fields[0][0].startswith("Deletion"): fieldCount -= 1
         if recordList:
-            [record.append(recordList[i]) for i in range(fieldCount)]
+            record = [recordList[i] for i in range(fieldCount)]
         elif recordDict:
             for field in self.fields:
                 if field[0] in recordDict:
@@ -1004,8 +1046,10 @@ class Writer:
                         record.append("")
                     else:
                         record.append(val)
-        if record:
-            self.records.append(record)
+        else:
+            # Blank fields for empty record
+            record = ["" for i in range(fieldCount)]
+        self.records.append(record)
 
     def shape(self, i):
         return self._shapes[i]
@@ -1018,8 +1062,9 @@ class Writer:
         """Save an shp file."""
         if not hasattr(target, "write"):
             target = os.path.splitext(target)[0] + '.shp'
-        if not self.shapeType:
-            self.shapeType = self._shapes[0].shapeType
+        if self.shapeType is None:
+            # autoset file type to first non-null geometry
+            self.shapeType = next((s.shapeType for s in self._shapes if s.shapeType), NULL)
         self.shp = self.__getFileObj(target)
         self.__shapefileHeader(self.shp, headerType='shp')
         self.__shpRecords()
@@ -1028,8 +1073,9 @@ class Writer:
         """Save an shx file."""
         if not hasattr(target, "write"):
             target = os.path.splitext(target)[0] + '.shx'
-        if not self.shapeType:
-            self.shapeType = self._shapes[0].shapeType
+        if self.shapeType is None:
+            # autoset file type to first non-null geometry
+            self.shapeType = next((s.shapeType for s in self._shapes if s.shapeType != NULL), NULL)
         self.shx = self.__getFileObj(target)
         self.__shapefileHeader(self.shx, headerType='shx')
         self.__shxRecords()
@@ -1072,6 +1118,7 @@ class Writer:
             self.dbf.close()
             if generated:
                 return target
+            
 class Editor(Writer):
     def __init__(self, shapefile=None, shapeType=POINT, autoBalance=1):
         self.autoBalance = autoBalance
@@ -1095,16 +1142,19 @@ class Editor(Writer):
         """Deletes the specified part of any shape by specifying a shape
         number, part number, or point number."""
         # shape, part, point
-        if shape and part and point:
+        shape_param_exists = shape is not None
+        part_param_exists = part is not None
+        point_param_exists = point is not None
+        if shape_param_exists and part_param_exists and point_param_exists:
             del self._shapes[shape][part][point]
         # shape, part
-        elif shape and part and not point:
+        elif shape_param_exists and part_param_exists and not point_param_exists:
             del self._shapes[shape][part]
         # shape
-        elif shape and not part and not point:
+        elif shape_param_exists and not part_param_exists and not point_param_exists:
             del self._shapes[shape]
         # point
-        elif not shape and not part and point:
+        elif not shape_param_exists and not part_param_exists and point_param_exists:
             for s in self._shapes:
                 if s.shapeType == 1:
                     del self._shapes[point]
@@ -1112,11 +1162,11 @@ class Editor(Writer):
                     for part in s.parts:
                         del s[part][point]
         # part, point
-        elif not shape and part and point:
+        elif not shape_param_exists and part_param_exists and point_param_exists:
             for s in self._shapes:
                 del s[part][point]
         # part
-        elif not shape and part and not point:
+        elif not shape_param_exists and part_param_exists and not point_param_exists:
             for s in self._shapes:
                 del s[part]
 
@@ -1190,11 +1240,17 @@ class Editor(Writer):
         fieldName.replace(' ', '_')
 
 # Begin Testing
-def test():
+def test(**kwargs):
     import doctest
     doctest.NORMALIZE_WHITESPACE = 1
-    doctest.testfile("README.md", verbose=1)
-
+    verbosity = kwargs.get('verbose', 0)
+    if verbosity == 0:
+        print('Running doctests...')
+    failure_count, test_count = doctest.testfile("README.md", verbose=verbosity)
+    if verbosity == 0 and failure_count == 0:
+        print('All test passed successfully')
+    return failure_count
+    
 if __name__ == "__main__":
     """
     Doctests are contained in the file 'README.md'. This library was originally developed
@@ -1202,4 +1258,5 @@ if __name__ == "__main__":
     testing libraries but for now unit testing is done using what's available in
     2.3.
     """
-    test()
+    failure_count = test()
+    sys.exit(failure_count)
